import os
import numpy as np
import pytest

from nptdms import TdmsFile
from nptdms.test.util import (
    GeneratedFile,
    hexlify_value,
    string_hexlify,
    segment_objects_metadata,
    channel_metadata,
    channel_metadata_with_no_data,
    channel_metadata_with_repeated_structure)
from nptdms.test.scenarios import TDS_TYPE_INT32, timestamp_data_chunk


@pytest.mark.benchmark(group='read-all-data')
def test_read_contiguous_data(benchmark):
    """ Benchmark reading a file with multiple channels of contiguous data
    """
    tdms_file = benchmark(read_from_start, get_contiguous_file().get_bytes_io_file())

    np.testing.assert_equal(tdms_file['group']['channel1'][:], np.repeat([1], 10000))
    np.testing.assert_equal(tdms_file['group']['channel2'][:], np.repeat([2], 10000))
    np.testing.assert_equal(tdms_file['group']['channel3'][:], np.repeat([3], 10000))
    np.testing.assert_equal(tdms_file['group']['channel4'][:], np.repeat([4], 10000))


@pytest.mark.benchmark(group='read-all-data')
def test_read_interleaved_data(benchmark):
    """ Benchmark reading a file with interleaved data
    """
    tdms_file = benchmark(read_from_start, get_interleaved_file().get_bytes_io_file())

    np.testing.assert_equal(tdms_file['group']['channel1'][:], np.repeat([1], 10000))
    np.testing.assert_equal(tdms_file['group']['channel2'][:], np.repeat([2], 10000))
    np.testing.assert_equal(tdms_file['group']['channel3'][:], np.repeat([3], 10000))
    np.testing.assert_equal(tdms_file['group']['channel4'][:], np.repeat([4], 10000))


@pytest.mark.benchmark(group='read-all-channel')
def test_read_contiguous_data_channel(benchmark):
    """ Benchmark reading a single channel's data from a contiguous data file
    """
    with TdmsFile.open(get_contiguous_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = benchmark(read_channel_data, channel)

        expected_data = np.repeat([3], 10000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-all-channel')
def test_read_interleaved_data_channel(benchmark):
    """ Benchmark reading a single channel's data from an interleaved data file
    """
    with TdmsFile.open(get_interleaved_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = benchmark(read_channel_data, channel)

        expected_data = np.repeat([3], 10000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-all-channel')
def test_stream_contiguous_data_channel(benchmark):
    """ Benchmark streaming channel data from a contiguous data file
    """
    with TdmsFile.open(get_contiguous_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = benchmark(stream_chunks, channel)

        channel_data = np.concatenate(channel_data)
        expected_data = np.repeat([3], 10000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-all-channel')
def test_stream_interleaved_data_channel(benchmark):
    """ Benchmark streaming channel data from an interleaved data file
    """
    with TdmsFile.open(get_interleaved_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = benchmark(stream_chunks, channel)

        channel_data = np.concatenate(channel_data)
        expected_data = np.repeat([3], 10000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='slice-channel')
def test_slice_contiguous_data_channel(benchmark):
    """ Benchmark reading a slice of data from a contiguous data file
    """
    with TdmsFile.open(get_contiguous_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = benchmark(get_slice, channel, 5555, 6555)

        expected_data = np.repeat([3], 1000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='slice-channel')
def test_slice_interleaved_data_channel(benchmark):
    """ Benchmark reading a slice of data from an interleaved data file
    """
    with TdmsFile.open(get_interleaved_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = benchmark(get_slice, channel, 5555, 6555)

        expected_data = np.repeat([3], 1000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-all-channel')
def test_index_contiguous_data_channel(benchmark):
    """ Benchmark reading a data from a contiguous data file using integer indices
    """
    with TdmsFile.open(get_contiguous_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = np.zeros(10000, dtype=channel.dtype)
        benchmark(index_values, channel, channel_data)

        expected_data = np.repeat([3], 10000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-all-channel')
def test_index_interleaved_data_channel(benchmark):
    """ Benchmark reading a data from a interleaved data file using integer indices
    """
    with TdmsFile.open(get_interleaved_file().get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel3']
        channel_data = np.zeros(10000, dtype=channel.dtype)
        benchmark(index_values, channel, channel_data)

        expected_data = np.repeat([3], 10000)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-scaled-channel')
def test_stream_scaled_data_chunks(benchmark):
    """ Benchmark streaming channel data when the data is scaled
    """
    properties = {
        "NI_Number_Of_Scales":
            (3, "01 00 00 00"),
        "NI_Scale[0]_Scale_Type":
            (0x20, hexlify_value("<I", len("Linear")) + string_hexlify("Linear")),
        "NI_Scale[0]_Linear_Slope":
            (10, hexlify_value("<d", 2.0)),
        "NI_Scale[0]_Linear_Y_Intercept":
            (10, hexlify_value("<d", 10.0))
    }
    test_file = GeneratedFile()
    data_array = np.arange(0, 1000, dtype=np.dtype('int32'))
    data = data_array.tobytes()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 100, properties),
        ),
        data, binary_data=True
    )
    for _ in range(0, 9):
        test_file.add_segment(
            ("kTocRawData", ), "", data, binary_data=True)

    with TdmsFile.open(test_file.get_bytes_io_file()) as tdms_file:
        channel = tdms_file['group']['channel1']
        channel_data = benchmark(stream_chunks, channel)

        channel_data = np.concatenate(channel_data)
        expected_data = np.tile(10.0 + 2.0 * data_array, 10)
        np.testing.assert_equal(channel_data, expected_data)


@pytest.mark.benchmark(group='read-timestamp-data')
def test_read_timestamp_data(benchmark):
    """ Benchmark reading a file with timestamp data
    """
    timestamps = np.tile(np.array([
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
        np.datetime64('2012-08-23T12:00:00.0', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
    ]), 200)
    data = timestamp_data_chunk(timestamps)

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 0x44, 200, {}),
        ),
        data
    )

    tdms_file = benchmark(read_from_start, test_file.get_bytes_io_file())

    np.testing.assert_equal(tdms_file['group']['channel1'][:], timestamps)


@pytest.mark.benchmark(group='read-timestamp-data')
def test_read_interleaved_timestamp_data(benchmark):
    """ Benchmark reading a file with interleaved timestamp data
    """
    timestamps = np.tile(np.array([
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
        np.datetime64('2012-08-23T12:00:00.0', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
        np.datetime64('2012-08-23T00:00:00.123', 'us'),
        np.datetime64('2012-08-23T01:02:03.456', 'us'),
        np.datetime64('2012-08-23T12:00:00.0', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
        np.datetime64('2012-08-23T12:02:03.9999', 'us'),
    ]), 100)
    data = timestamp_data_chunk(timestamps)

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 0x44, 100, {}),
            channel_metadata("/'group'/'channel2'", 0x44, 100, {}),
        ),
        data
    )

    tdms_file = benchmark(read_from_start, test_file.get_bytes_io_file())

    np.testing.assert_equal(tdms_file['group']['channel1'][:], timestamps[0::2])
    np.testing.assert_equal(tdms_file['group']['channel2'][:], timestamps[1::2])


@pytest.mark.benchmark(group='read-metadata')
def test_complex_metadata_reading(benchmark):
    """ Benchmark reading metadata for a file with many channels and segments with alternating sets of objects
    """
    test_file = GeneratedFile()
    data = np.array([0] * 5, dtype=np.dtype('int32')).tobytes()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel0'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel4'", TDS_TYPE_INT32, 1),
        ),
        data, binary_data=True
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel5'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel6'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel7'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel8'", TDS_TYPE_INT32, 1),
            channel_metadata("/'group'/'channel9'", TDS_TYPE_INT32, 1),
        ),
        data, binary_data=True
    )
    for _ in range(9):
        test_file.add_segment(
            ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
            segment_objects_metadata(
                channel_metadata_with_no_data("/'group'/'channel0'"),
                channel_metadata_with_no_data("/'group'/'channel1'"),
                channel_metadata_with_no_data("/'group'/'channel2'"),
                channel_metadata_with_no_data("/'group'/'channel3'"),
                channel_metadata_with_no_data("/'group'/'channel4'"),
                channel_metadata_with_repeated_structure("/'group'/'channel5'"),
                channel_metadata_with_repeated_structure("/'group'/'channel6'"),
                channel_metadata_with_repeated_structure("/'group'/'channel7'"),
                channel_metadata_with_repeated_structure("/'group'/'channel8'"),
                channel_metadata_with_repeated_structure("/'group'/'channel9'"),
            ),
            data, binary_data=True
        )
        test_file.add_segment(
            ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
            segment_objects_metadata(
                channel_metadata_with_repeated_structure("/'group'/'channel0'"),
                channel_metadata_with_repeated_structure("/'group'/'channel1'"),
                channel_metadata_with_repeated_structure("/'group'/'channel2'"),
                channel_metadata_with_repeated_structure("/'group'/'channel3'"),
                channel_metadata_with_repeated_structure("/'group'/'channel4'"),
                channel_metadata_with_no_data("/'group'/'channel5'"),
                channel_metadata_with_no_data("/'group'/'channel6'"),
                channel_metadata_with_no_data("/'group'/'channel7'"),
                channel_metadata_with_no_data("/'group'/'channel8'"),
                channel_metadata_with_no_data("/'group'/'channel9'"),
            ),
            data, binary_data=True
        )

    tdms_file = benchmark(read_metadata_from_start, test_file.get_bytes_io_file())

    assert len(tdms_file) == 1
    assert len(tdms_file['group']) == 10
    for channel_num in range(10):
        assert len(tdms_file['group']['channel{0}'.format(channel_num)]) == 10


def get_contiguous_file():
    test_file = GeneratedFile()
    data_chunk = np.repeat(np.array([1, 2, 3, 4], dtype=np.dtype('int32')), 100)
    data_array = np.tile(data_chunk, 10)
    data = data_array.tobytes()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 100),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 100),
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 100),
            channel_metadata("/'group'/'channel4'", TDS_TYPE_INT32, 100),
        ),
        data, binary_data=True
    )
    for _ in range(0, 9):
        test_file.add_segment(
            ("kTocRawData", ), "", data, binary_data=True)
    return test_file


def get_interleaved_file():
    test_file = GeneratedFile()
    data_array = np.tile(np.array([1, 2, 3, 4], dtype=np.dtype('int32')), 1000)
    data = data_array.tobytes()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 100),
            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 100),
            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 100),
            channel_metadata("/'group'/'channel4'", TDS_TYPE_INT32, 100),
        ),
        data, binary_data=True
    )
    for _ in range(0, 9):
        test_file.add_segment(
            ("kTocRawData", "kTocInterleavedData"), "", data, binary_data=True)
    return test_file


def read_from_start(file):
    file.seek(0, os.SEEK_SET)
    return TdmsFile.read(file)


def read_metadata_from_start(file):
    file.seek(0, os.SEEK_SET)
    return TdmsFile.read_metadata(file)


def read_channel_data(chan):
    return chan[:]


def stream_chunks(chan):
    all_data = []
    for chunk in chan.data_chunks():
        all_data.append(chunk[:])
    return all_data


def get_slice(chan, start, stop):
    return chan[start:stop]


def index_values(chan, target):
    for i in range(len(chan)):
        target[i] = chan[i]



"""Test reading of example TDMS files"""

from collections import defaultdict
import logging
import os
import sys
from shutil import copyfile
import tempfile
import weakref
from hypothesis import (assume, given, example, settings, strategies, HealthCheck)
import numpy as np
import pytest
from nptdms import TdmsFile
from nptdms.log import log_manager
from nptdms.test.util import (
    BytesIoTestFile,
    GeneratedFile,
    basic_segment,
    channel_metadata,
    compare_arrays,
    hexlify_value,
    segment_objects_metadata,
    string_hexlify,
)
from nptdms.test import scenarios


@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
def test_read_channel_data(test_file, expected_data):
    """Test reading data"""

    with test_file.get_tempfile() as temp_file:
        tdms_data = TdmsFile.read(temp_file.file)

    for ((group, channel), expected_data) in expected_data.items():
        channel_obj = tdms_data[group][channel]
        actual_data = channel_obj.data
        assert actual_data.dtype == expected_data.dtype
        assert channel_obj.dtype == expected_data.dtype
        compare_arrays(actual_data, expected_data)


@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
def test_lazily_read_channel_data(test_file, expected_data):
    """Test reading channel data lazily"""

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_data) in expected_data.items():
                actual_data = tdms_file[group][channel].read_data()
                assert actual_data.dtype == expected_data.dtype
                compare_arrays(actual_data, expected_data)


def test_read_raw_channel_data():
    """Test reading raw channel data"""

    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    with test_file.get_tempfile() as temp_file:
        tdms_file = TdmsFile.read(temp_file.file)
        for ((group, channel), expected_data) in expected_data.items():
            actual_data = tdms_file[group][channel].read_data(scaled=False)
            assert actual_data.dtype == expected_data.dtype
            compare_arrays(actual_data, expected_data)


def test_lazily_read_raw_channel_data():
    """Test reading raw channel data lazily"""

    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_data) in expected_data.items():
                actual_data = tdms_file[group][channel].read_data(scaled=False)
                assert actual_data.dtype == expected_data.dtype
                compare_arrays(actual_data, expected_data)


def test_read_raw_channel_data_slice():
    """Test reading a slice of raw channel data"""

    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    with test_file.get_tempfile() as temp_file:
        tdms_file = TdmsFile.read(temp_file.file)
        for ((group, channel), expected_data) in expected_data.items():
            actual_data = tdms_file[group][channel].read_data(offset=1, length=2, scaled=False)
            assert actual_data.dtype == expected_data.dtype
            compare_arrays(actual_data, expected_data[1:3])


def test_lazily_read_raw_channel_data_slice():
    """Test reading raw channel data lazily"""

    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_data) in expected_data.items():
                actual_data = tdms_file[group][channel].read_data(offset=1, length=2, scaled=False)
                assert actual_data.dtype == expected_data.dtype
                compare_arrays(actual_data, expected_data[1:3])


def test_lazily_read_channel_data_with_file_path():
    """Test reading channel data lazily after initialising with a file path
    """
    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    temp_file = test_file.get_tempfile(delete=False)
    try:
        temp_file.file.close()
        with TdmsFile.open(temp_file.name) as tdms_file:
            for ((group, channel), expected_data) in expected_data.items():
                actual_data = tdms_file[group][channel].read_data()
                assert actual_data.dtype == expected_data.dtype
                compare_arrays(actual_data, expected_data)
    finally:
        os.remove(temp_file.name)


@given(offset=strategies.integers(0, 100), length=strategies.integers(0, 100))
@example(offset=0, length=0)
@example(offset=0, length=100)
@example(offset=0, length=5)
@example(offset=0, length=10)
def test_reading_subset_of_data(offset, length):
    channel_data = np.arange(0, 100, 1, dtype=np.int32)
    # Split data into different sized segments
    segment_data = [
        channel_data[0:10],
        channel_data[10:20],
        channel_data[20:60],
        channel_data[60:80],
        channel_data[80:90],
        channel_data[90:100],
    ]
    hex_segment_data = [
        "".join(hexlify_value('<i', x) for x in data) for data in segment_data]
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 5),
        ),
        hex_segment_data[0]
    )
    for hex_data in hex_segment_data[1:]:
        test_file.add_segment(("kTocRawData", ), "", hex_data)

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            channel_subset = tdms_file['group']['channel1'].read_data(offset, length)
            expected_data = channel_data[offset:offset + length]
            assert len(channel_subset) == len(expected_data)
            np.testing.assert_equal(channel_subset, expected_data)


@settings(suppress_health_check=[HealthCheck.filter_too_much])
@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
@given(offset=strategies.integers(0, 10), length=strategies.integers(0, 10))
def test_reading_subset_of_data_for_scenario(test_file, expected_data, offset, length):
    """Test reading a subset of a channel's data
    """
    assume(any(offset <= len(d) for d in expected_data.values()))
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_data) in expected_data.items():
                actual_data = tdms_file[group][channel].read_data(offset, length)
                compare_arrays(actual_data, expected_data[offset:offset + length])


@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
def test_stream_data_chunks(test_file, expected_data):
    """Test streaming chunks of data from a TDMS file
    """
    data_arrays = defaultdict(list)
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for chunk in tdms_file.data_chunks():
                for group in chunk.groups():
                    for channel in group.channels():
                        key = (group.name, channel.name)
                        assert channel.offset == len(data_arrays[key])
                        data_arrays[key].extend(channel[:])

    for ((group, channel), expected_data) in expected_data.items():
        actual_data = data_arrays[(group, channel)]
        compare_arrays(actual_data, expected_data)


def test_indexing_and_iterating_data_chunks():
    """Test streaming chunks of data from a TDMS file and indexing into chunks
    """
    test_file, expected_data = scenarios.single_segment_with_two_channels().values
    data_arrays = defaultdict(list)
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for chunk in tdms_file.data_chunks():
                for (group, channel) in expected_data.keys():
                    key = (group, channel)
                    channel_chunk = chunk[group][channel]
                    data_arrays[key].extend(list(channel_chunk))

    for ((group, channel), expected_data) in expected_data.items():
        actual_data = data_arrays[(group, channel)]
        compare_arrays(actual_data, expected_data)


@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
def test_stream_channel_data_chunks(test_file, expected_data):
    """Test streaming chunks of data for a single channel from a TDMS file
    """
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                actual_data = []
                for chunk in tdms_file[group][channel].data_chunks():
                    assert chunk.offset == len(actual_data)
                    actual_data.extend(chunk[:])
                compare_arrays(actual_data, expected_channel_data)


def test_iterate_channel_data_in_open_mode():
    """Test iterating over channel data after opening a file without reading data
    """
    test_file, expected_data = scenarios.chunked_segment().values

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                actual_data = []
                for value in tdms_file[group][channel]:
                    actual_data.append(value)
                compare_arrays(actual_data, expected_channel_data)


def test_iterate_channel_data_in_read_mode():
    """Test iterating over channel data after reading all data
    """
    test_file, expected_data = scenarios.chunked_segment().values

    with test_file.get_tempfile() as temp_file:
        tdms_file = TdmsFile.read(temp_file.file)
        for ((group, channel), expected_channel_data) in expected_data.items():
            actual_data = []
            for value in tdms_file[group][channel]:
                actual_data.append(value)
            compare_arrays(actual_data, expected_channel_data)


def test_iterate_file_and_groups():
    """ Test iterating over TdmsFile and TdmsGroup uses key values
    """
    test_file, expected_data = scenarios.chunked_segment().values

    with test_file.get_tempfile() as temp_file:
        tdms_file = TdmsFile.read(temp_file.file)
        for group_name in tdms_file:
            group = tdms_file[group_name]
            for channel_name in group:
                channel = group[channel_name]
                expected_channel_data = expected_data[(group_name, channel_name)]
                compare_arrays(channel.data, expected_channel_data)


def test_indexing_channel_after_read_data():
    """ Test indexing into a channel after reading all data
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        tdms_file = TdmsFile.read(temp_file.file)
    for ((group, channel), expected_channel_data) in expected_data.items():
        channel_object = tdms_file[group][channel]
        assert channel_object[0] == expected_channel_data[0]
        compare_arrays(channel_object[:], expected_channel_data)


@given(index=strategies.integers(0, 7))
def test_indexing_channel_with_integer(index):
    """ Test indexing into a channel with an integer index
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                assert channel_object[index] == expected_channel_data[index]


@given(index=strategies.integers(-8, -1))
def test_indexing_channel_with_negative_integer(index):
    """ Test indexing into a channel with a negative integer index
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                assert channel_object[index] == expected_channel_data[index]


@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
def test_indexing_channel_with_integer_and_caching(test_file, expected_data):
    """ Test indexing into a channel with an integer index, reusing the same file to test caching
    """
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                values = []
                for i in range(len(channel_object)):
                    values.append(channel_object[i])
                compare_arrays(values, expected_channel_data)


@given(index=strategies.integers(0, 1))
def test_indexing_timestamp_channel_with_integer(index):
    """ Test indexing into a timestamp data channel with an integer index
    """
    test_file, expected_data = scenarios.timestamp_data().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                assert channel_object[index] == expected_channel_data[index]


def test_indexing_scaled_channel_with_integer():
    """ Test indexing into a channel with an integer index when the channel is scaled
    """
    test_file, expected_data = scenarios.scaled_data().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                values = []
                for i in range(len(channel_object)):
                    values.append(channel_object[i])
                compare_arrays(values, expected_channel_data)


def test_indexing_channel_with_ellipsis():
    """ Test indexing into a channel with ellipsis returns all data
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                compare_arrays(channel_object[...], expected_channel_data)


@pytest.fixture(scope="module")
def opened_tdms_file():
    """ Allow re-use of an opened TDMS file
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            yield tdms_file, expected_data


@given(
    start=strategies.integers(-10, 10) | strategies.none(),
    stop=strategies.integers(-10, 10) | strategies.none(),
    step=strategies.integers(-5, 5).filter(lambda i: i != 0) | strategies.none(),
)
@settings(max_examples=1000)
def test_indexing_channel_with_slice(opened_tdms_file, start, stop, step):
    """ Test indexing into a channel with a slice
    """
    tdms_file, expected_data = opened_tdms_file
    for ((group, channel), expected_channel_data) in expected_data.items():
        channel_object = tdms_file[group][channel]
        compare_arrays(channel_object[start:stop:step], expected_channel_data[start:stop:step])


@pytest.mark.parametrize('index', [-9, 8])
def test_indexing_channel_with_invalid_integer_raises_error(index):
    """ Test indexing into a channel with an invalid integer index
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                with pytest.raises(IndexError):
                    _ = channel_object[index]


def test_indexing_channel_with_zero_step_raises_error():
    """ Test indexing into a channel with a slice with zero step size raises an error
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                with pytest.raises(ValueError) as exc_info:
                    _ = channel_object[::0]
                assert str(exc_info.value) == "Step size cannot be zero"


@pytest.mark.parametrize('index', ["test", None])
def test_indexing_channel_with_invalid_type_raises_error(index):
    """ Test indexing into a channel with an invalid index type
    """
    test_file, expected_data = scenarios.chunked_segment().values
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            for ((group, channel), expected_channel_data) in expected_data.items():
                channel_object = tdms_file[group][channel]
                with pytest.raises(TypeError) as exc_info:
                    _ = channel_object[index]
                assert "Invalid index type" in str(exc_info.value)


def test_invalid_offset_in_read_data_throws():
    """ Exception is thrown when reading a subset of data with an invalid offset
    """
    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    group, channel = list(expected_data.keys())[0]
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            with pytest.raises(ValueError) as exc_info:
                tdms_file[group][channel].read_data(-1, 5)
            assert "offset must be non-negative" in str(exc_info.value)


def test_invalid_length_in_read_data_throws():
    """ Exception is thrown when reading a subset of data with an invalid length
    """
    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    group, channel = list(expected_data.keys())[0]
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            with pytest.raises(ValueError) as exc_info:
                tdms_file[group][channel].read_data(0, -5)
            assert "length must be non-negative" in str(exc_info.value)


def test_read_data_after_close_throws():
    """ Trying to read after opening and closing without reading data should throw
    """
    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    group, channel = list(expected_data.keys())[0]
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            pass
        with pytest.raises(RuntimeError) as exc_info:
            tdms_file[group][channel].read_data()
        assert "Cannot read data after the underlying TDMS reader is closed" in str(exc_info.value)


def test_access_data_property_after_opening_throws():
    """ Accessing the data property after opening without reading data should throw
    """
    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    group, channel = list(expected_data.keys())[0]
    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            with pytest.raises(RuntimeError) as exc_info:
                _ = tdms_file[group][channel].data
            assert "Channel data has not been read" in str(exc_info.value)

            with pytest.raises(RuntimeError) as exc_info:
                _ = tdms_file[group][channel].raw_data
            assert "Channel data has not been read" in str(exc_info.value)

            with pytest.raises(RuntimeError) as exc_info:
                _ = tdms_file[group][channel].raw_scaler_data
            assert "Channel data has not been read" in str(exc_info.value)


@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
def test_read_with_index_file(test_file, expected_data):
    """ Test reading a file with an associated tdms_index file
    """
    with test_file.get_tempfile_with_index() as tdms_file_path:
        tdms_file = TdmsFile.read(tdms_file_path)

    for ((group, channel), expected_channel_data) in expected_data.items():
        channel_obj = tdms_file[group][channel]
        compare_arrays(channel_obj.data, expected_channel_data)


@pytest.mark.skipif(sys.version_info < (3, 4), reason="pathlib only available in stdlib since 3.4")
def test_read_file_passed_as_pathlib_path():
    """ Test reading a file when using a pathlib Path object
    """
    import pathlib

    test_file, expected_data = scenarios.single_segment_with_one_channel().values

    with test_file.get_tempfile_with_index() as tdms_file_path_str:
        tdms_file_path = pathlib.Path(tdms_file_path_str)
        tdms_file = TdmsFile.read(tdms_file_path)

    for ((group, channel), expected_channel_data) in expected_data.items():
        channel_obj = tdms_file[group][channel]
        compare_arrays(channel_obj.data, expected_channel_data)


def test_read_with_mismatching_index_file():
    """ Test that reading data when the index file doesn't match the data file raises an error
    """

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 2),
            channel_metadata("/'group'/'channel2'", 3, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 2),
            channel_metadata("/'group'/'channel2'", 3, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )

    test_file_with_index = GeneratedFile()
    test_file_with_index.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 3),
            channel_metadata("/'group'/'channel2'", 3, 3),
        ),
        "01 00 00 00" "02 00 00 00" "03 00 00 00"
        "04 00 00 00" "05 00 00 00" "06 00 00 00"
    )
    test_file_with_index.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 3),
            channel_metadata("/'group'/'channel2'", 3, 3),
        ),
        "01 00 00 00" "02 00 00 00" "03 00 00 00"
        "04 00 00 00" "05 00 00 00" "06 00 00 00"
    )

    with test_file.get_tempfile(delete=False) as tdms_file:
        with test_file_with_index.get_tempfile_with_index() as tdms_file_with_index_path:
            # Move index file from second file to match the name of the first file
            new_index_file = tdms_file.name + '_index'
            copyfile(tdms_file_with_index_path + '_index', new_index_file)
            try:
                tdms_file.file.close()
                with pytest.raises(ValueError) as exc_info:
                    _ = TdmsFile.read(tdms_file.name)
                assert 'Check that the tdms_index file matches the tdms data file' in str(exc_info.value)
            finally:
                os.remove(new_index_file)
                os.remove(tdms_file.name)


def test_get_len_of_file():
    """Test getting the length of a TdmsFile
    """
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    assert len(tdms_data) == 1


def test_get_len_of_group():
    """Test getting the length of a TdmsGroup
    """
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    assert len(tdms_data['Group']) == 2


def test_key_error_getting_invalid_group():
    """Test getting a group that doesn't exist raises a KeyError
    """
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    with pytest.raises(KeyError) as exc_info:
        _ = tdms_data['non-existent group']
    assert 'non-existent group' in str(exc_info.value)


def test_key_error_getting_invalid_channel():
    """Test getting a channel that doesn't exist raises a KeyError
    """
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    group = tdms_data['Group']
    with pytest.raises(KeyError) as exc_info:
        _ = group['non-existent channel']
    assert 'non-existent channel' in str(exc_info.value)
    assert 'Group' in str(exc_info.value)


def test_group_property_read():
    """Test reading property of a group"""

    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    group = tdms_data["Group"]
    assert group.properties["num"] == 10


def test_time_track():
    """Add a time track to waveform data"""

    test_file = GeneratedFile()
    (toc, metadata, data) = basic_segment()
    test_file.add_segment(toc, metadata, data)
    tdms_data = test_file.load()

    channel = tdms_data["Group"]["Channel2"]
    time = channel.time_track()
    assert len(time) == len(channel.data)
    epsilon = 1.0E-15
    assert abs(time[0]) < epsilon
    assert abs(time[1] - 0.1) < epsilon


def test_memmapped_read():
    """Test reading data into memmapped arrays"""

    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load(memmap_dir=tempfile.gettempdir())

    data = tdms_data["Group"]["Channel1"].data
    assert len(data) == 2
    assert data[0] == 1
    assert data[1] == 2
    data = tdms_data["Group"]["Channel2"].data
    assert len(data) == 2
    assert data[0] == 3
    assert data[1] == 4


def test_string_data():
    """Test reading a file with string data"""

    strings = ["abcdefg", "qwertyuiop"]

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "01 00 00 00"
        # Length of the object path
        "18 00 00 00")
    metadata += string_hexlify("/'Group'/'StringChannel'")
    metadata += (
        # Length of index information
        "1C 00 00 00"
        # Raw data data type
        "20 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of bytes in data
        "19 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")
    data = (
        "07 00 00 00"  # index to after first string
        "11 00 00 00"  # index to after second string
    )
    for string in strings:
        data += string_hexlify(string)
    test_file.add_segment(toc, metadata, data)
    tdms_data = test_file.load()

    channel = tdms_data["Group"]["StringChannel"]
    assert len(channel.data) == len(strings)
    assert channel.data.dtype == channel.dtype
    for expected, read in zip(strings, channel.data):
        assert expected == read


def test_string_data_in_interleaved_segment():
    """ Test reading a file with string data where a segment
        has the interleaved data flag set but only one channel.
        (See issue #239)
    """

    strings = ["abcdefg", "qwertyuiop"]

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
    metadata = (
        "01 00 00 00"  # Number of objects
        "18 00 00 00")  # Length of the object path
    metadata += string_hexlify("/'Group'/'StringChannel'")
    metadata += (
        "1C 00 00 00"  # Length of index information
        "20 00 00 00"  # Raw data data type
        "01 00 00 00"  # Dimension
        "02 00 00 00 00 00 00 00"  # Number of raw data values
        "19 00 00 00 00 00 00 00"  # Number of bytes in data
        "00 00 00 00")  # Number of properties (0)
    data = (
        "07 00 00 00"  # index to after first string
        "11 00 00 00"  # index to after second string
    )
    for string in strings:
        data += string_hexlify(string)
    test_file.add_segment(toc, metadata, data)
    tdms_data = test_file.load()

    channel = tdms_data["Group"]["StringChannel"]
    assert len(channel.data) == len(strings)
    assert channel.data.dtype == channel.dtype
    for expected, read in zip(strings, channel.data):
        assert expected == read


def test_incomplete_segment_with_string_data():
    """ Test incomplete last segment, eg. if LabView crashed, with string data
    """
    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "01 00 00 00"
        # Length of the object path
        "18 00 00 00")
    metadata += string_hexlify("/'Group'/'StringChannel'")
    metadata += (
        # Length of index information
        "1C 00 00 00"
        # Raw data data type
        "20 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of bytes in data
        "19 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")
    data = "00 00 00 00"
    test_file.add_segment(toc, metadata, data, incomplete=True)

    tdms_data = test_file.load()

    channel = tdms_data["Group"]["StringChannel"]
    assert len(channel) == 0


def test_slash_and_space_in_name():
    """Test name like '01/02/03 something'"""

    group_1 = "01/02/03 something"
    channel_1 = "04/05/06 another thing"
    group_2 = "01/02/03 a"
    channel_2 = "04/05/06 b"

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'{0}'/'{1}'".format(group_1, channel_1), 3, 2),
            channel_metadata("/'{0}'/'{1}'".format(group_2, channel_2), 3, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "03 00 00 00" "04 00 00 00"
    )

    tdms_data = test_file.load()

    assert len(tdms_data.groups()) == 2
    assert len(tdms_data[group_1].channels()) == 1
    assert len(tdms_data[group_2].channels()) == 1
    data_1 = tdms_data[group_1][channel_1].data
    assert len(data_1) == 2
    data_2 = tdms_data[group_2][channel_2].data
    assert len(data_2) == 2


def test_single_quote_in_name():
    group_1 = "group''s name"
    channel_1 = "channel''s name"

    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'{0}'/'{1}'".format(group_1, channel_1), 3, 2),
        ),
        "01 00 00 00" "02 00 00 00"
    )

    tdms_data = test_file.load()

    assert len(tdms_data.groups()) == 1
    assert len(tdms_data["group's name"].channels()) == 1
    data_1 = tdms_data["group's name"]["channel's name"].data
    assert len(data_1) == 2


def test_group_object_paths():
    """Test the path and name properties for a group"""
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    group = tdms_data["Group"]
    assert group.path == "/'Group'"
    assert group.name == "Group"


def test_channel_object_paths():
    """Test the path and name properties for a channel"""
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    channel = tdms_data["Group"]["Channel1"]
    assert channel.path == "/'Group'/'Channel1'"
    assert channel.name == "Channel1"
    assert channel.group_name == "Group"


def test_object_repr():
    """Test getting object representations of groups and channels
    """
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    group = tdms_data["Group"]
    assert repr(group) == "<TdmsGroup with path /'Group'>"

    channel = group["Channel1"]
    assert repr(channel) == "<TdmsChannel with path /'Group'/'Channel1'>"


def test_data_read_from_bytes_io():
    """Test reading data"""

    test_file = BytesIoTestFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    data = tdms_data["Group"]["Channel1"].data
    assert len(data) == 2
    assert data[0] == 1
    assert data[1] == 2
    data = tdms_data["Group"]["Channel2"].data
    assert len(data) == 2
    assert data[0] == 3
    assert data[1] == 4


def test_file_properties():
    """Test reading properties of the file (root object)"""

    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())

    tdms_file = test_file.load()

    file_props = tdms_file.properties
    assert file_props['num'] == 15


def test_debug_logging(caplog):
    """ Test loading a file with debug logging enabled
    """
    test_file, expected_data = scenarios.single_segment_with_one_channel().values

    log_manager.set_level(logging.DEBUG)
    _ = test_file.load()

    assert "Reading metadata for object /'group'/'channel1' with index header 0x00000014" in caplog.text
    assert "Object data type: Int32" in caplog.text


def test_memory_released_when_tdms_file_out_of_scope():
    """ Tests that when a TDMS file object goes out of scope,
        TDMS channels and their data are also freed.
        This ensures there are no circular references between a TDMS file
        and its channels, which would mean the GC is needed to free these objects.
    """

    test_file, expected_data = scenarios.single_segment_with_one_channel().values
    with test_file.get_tempfile() as temp_file:
        tdms_data = TdmsFile.read(temp_file.file)
        chan = tdms_data['group']['channel1']
        chan_ref = weakref.ref(chan)
        data_ref = weakref.ref(chan.data)
        raw_data_ref = weakref.ref(chan.raw_data)
    del tdms_data
    del chan

    assert raw_data_ref() is None
    assert data_ref() is None
    assert chan_ref() is None


def test_close_after_read():
    test_file, _ = scenarios.single_segment_with_one_channel().values
    temp_file = test_file.get_tempfile(delete=False)
    try:
        temp_file.file.close()
        tdms_data = TdmsFile.read(temp_file.name)
        tdms_data.close()
    finally:
        os.remove(temp_file.name)


def test_multiple_close_after_open():
    test_file, _ = scenarios.single_segment_with_one_channel().values
    temp_file = test_file.get_tempfile(delete=False)
    try:
        temp_file.file.close()
        with TdmsFile.open(temp_file.name) as tdms_data:
            tdms_data.close()
        tdms_data.close()
    finally:
        os.remove(temp_file.name)


def test_interleaved_segment_different_length():
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 3),
            channel_metadata("/'group'/'channel2'", 3, 2),
        ),
        "01 00 00 00" "02 00 00 00"
        "01 00 00 00" "02 00 00 00"
        "01 00 00 00"
    )

    with pytest.raises(ValueError) as exc_info:
        _ = test_file.load()
    assert str(exc_info.value) == "Cannot read interleaved data with different chunk sizes"


def test_multiple_string_channels_with_interleaved_flag():
    """ Test reading a segment with the interleaved data flag set but multiple string channels throws an exception
    """

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
    metadata = "02 00 00 00"  # Number of objects
    for channel in ["Channel1", "Channel2"]:
        channel_path = "/'Group'/'%s'" % channel
        metadata += hexlify_value('<L', len(channel_path))
        metadata += string_hexlify(channel_path)
        metadata += (
            "1C 00 00 00"  # Length of index information
            "20 00 00 00"  # Raw data data type
            "01 00 00 00"  # Dimension
            "02 00 00 00 00 00 00 00"  # Number of raw data values
            "02 00 00 00 00 00 00 00"  # Number of bytes in data
            "00 00 00 00")  # Number of properties
    data = "01 00 00 00 02 00 00 00 03 00 00 00 04 00 00 00"
    test_file.add_segment(toc, metadata, data)
    with pytest.raises(ValueError) as exc_info:
        _ = test_file.load()
    assert str(exc_info.value) == "Cannot read interleaved segment containing channels with unsized types"


def test_interleaved_segment_with_sized_and_unsized_types():
    test_file = GeneratedFile()
    string_channel_metadata = (
        "18 00 00 00" +  # Length of the object path
        string_hexlify("/'group'/'StringChannel'") +
        "1C 00 00 00"  # Length of index information
        "20 00 00 00"  # Raw data data type
        "01 00 00 00"  # Dimension
        "02 00 00 00 00 00 00 00"  # Number of raw data values
        "02 00 00 00 00 00 00 00"  # Number of bytes in data
        "00 00 00 00")  # Number of properties (0)
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 2),
            string_channel_metadata,
        ),
        "01 00 00 00" "02 00 00 00"
        "01 00 00 00" "02 00 00 00"
    )

    with pytest.raises(ValueError) as exc_info:
        _ = test_file.load()
    assert str(exc_info.value) == "Cannot read interleaved segment containing channels with unsized types"


def test_no_new_obj_list_for_first_segment():
    """ Test when kTocNewObjList is not set in the first segment
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 4),
        ),
        "01 00 00 00" "02 00 00 00" "03 00 00 00" "04 00 00 00"
    )
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 4),
        ),
        "05 00 00 00" "06 00 00 00" "07 00 00 00" "08 00 00 00"
    )

    with test_file.get_tempfile() as temp_file:
        tdms_data = TdmsFile.read(temp_file.file)
    channel_data = tdms_data['group']['channel1']
    compare_arrays(channel_data[:], [1, 2, 3, 4, 5, 6, 7, 8])



""" Test exporting TDMS data to HDF
"""
import pytest
import numpy as np
try:
    import h5py
except ImportError:
    pytest.skip("Skipping HDF tests as h5py is not installed", allow_module_level=True)

from nptdms import TdmsFile
from nptdms.test.util import (
    GeneratedFile,
    basic_segment,
    channel_metadata,
    compare_arrays,
    hexlify_value,
    segment_objects_metadata,
    string_hexlify,
)
from nptdms.test import scenarios


def test_hdf_channel_data(tmp_path):
    """ Test basic conversion of channel data to HDF
    """
    test_file, expected_data = scenarios.single_segment_with_two_channels().values

    tdms_data = test_file.load()
    h5_path = tmp_path / 'h5_data_test.h5'
    h5 = tdms_data.as_hdf(h5_path)

    for ((group, channel), expected_data) in expected_data.items():
        h5_channel = h5[group][channel]
        assert h5_channel.dtype.kind == 'i'
        np.testing.assert_almost_equal(h5_channel[...], expected_data)
    h5.close()


def test_streaming_to_hdf(tmp_path):
    """ Test conversion of channel data to HDF when streaming data from disk
    """
    test_file, expected_data = scenarios.chunked_segment().values

    with test_file.get_tempfile() as temp_file:
        with TdmsFile.open(temp_file.file) as tdms_file:
            h5_path = tmp_path / 'h5_streaming_data_test.h5'
            h5 = tdms_file.as_hdf(h5_path)

    for ((group, channel), expected_data) in expected_data.items():
        h5_channel = h5[group][channel]
        np.testing.assert_almost_equal(h5_channel[...], expected_data)
    h5.close()


def test_int_data_types(tmp_path):
    """ Test conversion of signed and unsigned integer types to HDF
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'i8'", 1, 4),
            channel_metadata("/'group'/'u8'", 5, 4),
            channel_metadata("/'group'/'i16'", 2, 4),
            channel_metadata("/'group'/'u16'", 6, 4),
            channel_metadata("/'group'/'i32'", 3, 4),
            channel_metadata("/'group'/'u32'", 7, 4),
            channel_metadata("/'group'/'i64'", 4, 4),
            channel_metadata("/'group'/'u64'", 8, 4),
        ),
        "01 02 03 04"
        "01 02 03 04"
        "01 00 02 00 03 00 04 00"
        "01 00 02 00 03 00 04 00"
        "01 00 00 00 02 00 00 00 03 00 00 00 04 00 00 00"
        "01 00 00 00 02 00 00 00 03 00 00 00 04 00 00 00"
        "01 00 00 00 00 00 00 00 02 00 00 00 00 00 00 00 03 00 00 00 00 00 00 00 04 00 00 00 00 00 00 00"
        "01 00 00 00 00 00 00 00 02 00 00 00 00 00 00 00 03 00 00 00 00 00 00 00 04 00 00 00 00 00 00 00"
    )

    tdms_data = test_file.load()
    h5_path = tmp_path / 'h5_data_test.h5'
    h5 = tdms_data.as_hdf(h5_path)

    for chan, expected_dtype in [
            ('i8', np.dtype('int8')),
            ('u8', np.dtype('uint8')),
            ('i16', np.dtype('int16')),
            ('u16', np.dtype('uint16')),
            ('i32', np.dtype('int32')),
            ('u32', np.dtype('uint32')),
            ('i64', np.dtype('int64')),
            ('u64', np.dtype('uint64'))]:
        h5_channel = h5['group'][chan]
        assert h5_channel.dtype == expected_dtype
        np.testing.assert_almost_equal(h5_channel[...], [1, 2, 3, 4])
    h5.close()


def test_floating_point_data_types(tmp_path):
    """ Test conversion of f32 and f64 types to HDF
    """
    test_file = GeneratedFile()
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'f32'", 9, 4),
            channel_metadata("/'group'/'f64'", 10, 4),
        ),
        hexlify_value('<f', 1) +
        hexlify_value('<f', 2) +
        hexlify_value('<f', 3) +
        hexlify_value('<f', 4) +
        hexlify_value('<d', 1) +
        hexlify_value('<d', 2) +
        hexlify_value('<d', 3) +
        hexlify_value('<d', 4)
    )

    tdms_data = test_file.load()
    h5_path = tmp_path / 'h5_data_test.h5'
    h5 = tdms_data.as_hdf(h5_path)

    for chan, expected_dtype in [
            ('f32', np.dtype('float32')),
            ('f64', np.dtype('float64'))]:
        h5_channel = h5['group'][chan]
        assert h5_channel.dtype == expected_dtype
        np.testing.assert_almost_equal(h5_channel[...], [1.0, 2.0, 3.0, 4.0])
    h5.close()


def test_timestamp_data(tmp_path):
    """ Test conversion of timestamp channel data to HDF
        HDF doesn't support timestamps natively, so these are converted to strings
    """

    test_file, expected_data = scenarios.timestamp_data().values
    tdms_data = test_file.load()
    h5_path = tmp_path / 'h5_timestamp_test.h5'
    h5 = tdms_data.as_hdf(h5_path)

    for (group, channel), expected_values in expected_data.items():
        h5_channel = h5[group][channel]
        assert h5_channel.dtype.kind == 'S'
        expected_strings = np.datetime_as_string(expected_values, unit='us', timezone='UTC')
        expected_ascii = [s.encode('ascii') for s in expected_strings]
        compare_arrays(h5_channel[...], expected_ascii)
    h5.close()


def test_hdf_properties(tmp_path):
    """ Test properties are converted to attributes in HDF files
    """
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    tdms_data = test_file.load()

    h5_path = tmp_path / 'h5_properties_test.h5'
    h5 = tdms_data.as_hdf(h5_path)

    # File level properties
    assert h5.attrs['num'] == 15

    # Group properties
    assert h5['Group'].attrs['prop'] == 'value'
    assert h5['Group'].attrs['num'] == 10

    # Channel properties
    assert h5['Group']['Channel2'].attrs['wf_start_offset'] == 0.0
    assert h5['Group']['Channel2'].attrs['wf_increment'] == 0.1


def test_timestamp_property(tmp_path):
    """ Test a timestamp property is converted to an attribute in an HDF file
        HDF doesn't support timestamps natively, so these are converted to strings
    """
    test_file = GeneratedFile()
    properties = {
        "wf_start_time": (0x44, hexlify_value("<Q", 0) + hexlify_value("<q", 3524551547))
    }
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 2, properties),
        ),
        "01 00 00 00" "02 00 00 00"
    )
    tdms_data = test_file.load()

    h5_path = tmp_path / 'h5_properties_test.h5'
    h5 = tdms_data.as_hdf(h5_path)

    assert h5['group']['channel1'].attrs['wf_start_time'] == b'2015-09-08T10:05:47.000000Z'


def test_as_hdf_string(tmp_path):
    """ Test HDF5 conversion for string datatype
    """
    strings = ["abc123", "?<>~`!@#$%^&*()-=_+,.;'[]:{}|"]

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "01 00 00 00"
        # Length of the object path
        "11 00 00 00")
    metadata += string_hexlify("/'Group'/'String'")
    metadata += (
        # Length of index information
        "1C 00 00 00"
        # Raw data data type
        "20 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of bytes in data
        "2B 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00")
    data = (
        "06 00 00 00"  # index to after first string
        "24 00 00 00"  # index to after second string
    )
    for string in strings:
        data += string_hexlify(string)
    test_file.add_segment(toc, metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["String"].data
    assert len(data) == len(strings)
    for expected, read in zip(strings, data):
        assert expected == read

    h5_path = tmp_path / 'h5_strings_test.h5'
    h5 = tdms_data.as_hdf(h5_path)
    h5_strings = h5['Group']['String']
    assert h5_strings.dtype.kind == 'O'
    assert h5_strings.shape[0] == len(strings)
    for expected, read in zip(strings, h5_strings[...]):
        assert expected == _as_string(read)
    h5.close()


def test_unicode_string_data(tmp_path):
    """ Test HDF5 conversion for string datatype with non-ASCII data
    """
    strings = ["Hello, \u4E16\u754C", "\U0001F600"]
    sizes = [len(s.encode('utf-8')) for s in strings]

    test_file = GeneratedFile()
    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "01 00 00 00"
        # Length of the object path
        "11 00 00 00")
    metadata += string_hexlify("/'Group'/'String'")
    metadata += (
        # Length of index information
        "1C 00 00 00"
        # Raw data data type
        "20 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw data values
        "02 00 00 00"
        "00 00 00 00" +
        # Number of bytes in data, including index
        hexlify_value('q', sum(sizes) + 4 * len(sizes)) +
        # Number of properties (0)
        "00 00 00 00")
    data = ""
    offset = 0
    for size in sizes:
        # Index gives end positions of strings:
        offset += size
        data += hexlify_value('i', offset)
    for string in strings:
        data += string_hexlify(string)
    test_file.add_segment(toc, metadata, data)
    tdms_data = test_file.load()

    data = tdms_data["Group"]["String"].data
    assert len(data) == len(strings)
    for expected, read in zip(strings, data):
        assert expected == read

    h5_path = tmp_path / 'h5_unicode_strings_test.h5'
    h5 = tdms_data.as_hdf(h5_path)
    h5_strings = h5['Group']['String']
    assert h5_strings.dtype.kind == 'O'
    assert h5_strings.shape[0] == len(strings)
    for expected, read in zip(strings, h5_strings[...]):
        assert expected == _as_string(read)
    h5.close()


def test_add_to_file_under_group(tmp_path):
    """ Test adding TDMS data to an HDF file under a group
    """
    test_file, expected_data = scenarios.single_segment_with_two_channels().values
    preexisting_data = np.array([1.0, 2.0, 3.0])

    tdms_data = test_file.load()
    h5_path = tmp_path / 'h5_data_test.h5'
    h5 = h5py.File(h5_path, 'w')
    h5['preexisting_data'] = preexisting_data
    h5.close()

    h5 = tdms_data.as_hdf(h5_path, mode='a', group='tdms_data')

    for ((group, channel), expected_data) in expected_data.items():
        h5_channel = h5['tdms_data'][group][channel]
        np.testing.assert_almost_equal(h5_channel[...], expected_data)
    np.testing.assert_almost_equal(h5['preexisting_data'], preexisting_data)
    h5.close()


def _as_string(bytes_or_string):
    # h5py reads string data as bytes by default since 3.0
    if isinstance(bytes_or_string, bytes):
        return bytes_or_string.decode('UTF-8')
    return bytes_or_string



import numpy as np
from nptdms import types


def from_tdms_file(tdms_file, filepath, mode='w', group='/'):
    """
    Converts the TDMS file into an HDF5 file

    :param tdms_file: The TDMS file object to convert.
    :param filepath: The path of the HDF5 file you want to write to.
    :param mode: The write mode of the HDF5 file. This can be 'w' or 'a'
    :param group: A group in the HDF5 file that will contain the TDMS data.
    """
    import h5py

    # Groups in TDMS are mapped to the first level of the HDF5 hierarchy

    # Channels in TDMS are then mapped to the second level of the HDF5
    # hierarchy, under the appropriate groups.

    # Properties in TDMS are mapped to attributes in HDF5.
    # These all exist under the appropriate, channel group etc.

    h5file = h5py.File(filepath, mode)

    if group in h5file:
        container_group = h5file[group]
    else:
        container_group = h5file.create_group(group)

    # First write the properties at the root level
    for property_name, property_value in tdms_file.properties.items():
        container_group.attrs[property_name] = _hdf_attr_value(property_value)

    # Now iterate through groups and channels,
    # writing the properties and creating data sets
    datasets = {}
    for group in tdms_file.groups():
        # Write the group's properties
        container_group.create_group(group.name)
        for prop_name, prop_value in group.properties.items():
            container_group[group.name].attrs[prop_name] = _hdf_attr_value(prop_value)

        # Write properties and data for each channel
        for channel in group.channels():
            channel_key = group.name + '/' + channel.name

            if channel.data_type is types.String:
                # Encode as variable length UTF-8 strings
                datasets[channel.path] = container_group.create_dataset(
                    channel_key, (len(channel),), dtype=h5py.string_dtype())
            elif channel.data_type is types.TimeStamp:
                # Timestamps are represented as fixed length ASCII strings
                # because HDF doesn't natively support timestamps
                datasets[channel.path] = container_group.create_dataset(
                    channel_key, (len(channel),), dtype='S27')
            else:
                datasets[channel.path] = container_group.create_dataset(
                    channel_key, (len(channel),), dtype=channel.dtype)

            for prop_name, prop_value in channel.properties.items():
                container_group[channel_key].attrs[prop_name] = _hdf_attr_value(prop_value)

    # Set data
    if tdms_file.data_read:
        for group in tdms_file.groups():
            for channel in group.channels():
                datasets[channel.path][...] = _hdf_array(channel, channel.data)
    else:
        # Data hasn't been read into memory, stream it from disk
        for chunk in tdms_file.data_chunks():
            for group in chunk.groups():
                for channel_chunk in group.channels():
                    channel = tdms_file[group.name][channel_chunk.name]
                    offset = channel_chunk.offset
                    end = offset + len(channel_chunk)
                    datasets[channel.path][offset:end] = _hdf_array(channel, channel_chunk[:])

    return h5file


def _hdf_array(channel, data):
    """ Convert data array into a format suitable for initialising HDF data
    """
    if channel.data_type is types.TimeStamp:
        string_data = np.datetime_as_string(data, unit='us', timezone='UTC')
        return [s.encode('ascii') for s in string_data]
    return data


def _hdf_attr_value(value):
    """ Convert a value into a format suitable for an HDF attribute
    """
    if isinstance(value, np.datetime64):
        return np.string_(np.datetime_as_string(value, unit='us', timezone='UTC'))
    return value



import os
import sys
from nptdms import tdmsinfo
from nptdms.test.util import GeneratedFile, basic_segment
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


def test_tdmsinfo(capsys):
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    temp_file = test_file.get_tempfile(delete=False)
    try:
        temp_file.file.close()
        with patch.object(sys, 'argv', ['tdmsinfo.py', temp_file.name]):
            tdmsinfo.main()
            captured = capsys.readouterr()
            assert "/'Group'/'Channel1'" in captured.out
            assert "wf_start_offset" not in captured.out
    finally:
        os.remove(temp_file.name)


def test_tdmsinfo_with_properties(capsys):
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    temp_file = test_file.get_tempfile(delete=False)
    try:
        temp_file.file.close()
        with patch.object(sys, 'argv', ['tdmsinfo.py', temp_file.name, '--properties']):
            tdmsinfo.main()
            captured = capsys.readouterr()
            assert "/'Group'/'Channel1'" in captured.out
            assert "wf_start_offset: 0.0" in captured.out
            assert "length: 2" in captured.out
    finally:
        os.remove(temp_file.name)


def test_tdmsinfo_with_debug_output(caplog):
    test_file = GeneratedFile()
    test_file.add_segment(*basic_segment())
    temp_file = test_file.get_tempfile(delete=False)
    try:
        temp_file.file.close()
        with patch.object(sys, 'argv', ['tdmsinfo.py', temp_file.name, '--debug']):
            tdmsinfo.main()
            assert "Reading metadata for object /'Group'/'Channel1'" in caplog.text
    finally:
        os.remove(temp_file.name)



import os

from setuptools import setup


def read_version():
    here = os.path.abspath(os.path.dirname(__file__))
    version_path = os.path.sep.join((here, "nptdms", "version.py"))
    v_globals = {}
    v_locals = {}
    exec(open(version_path).read(), v_globals, v_locals)
    return v_locals['__version__']


setup(
  name = 'npTDMS',
  version = read_version(),
  description = ("Cross-platform, NumPy based module for reading "
    "TDMS files produced by LabView."),
  author = 'Adam Reeve',
  author_email = 'adreeve@gmail.com',
  url = 'https://github.com/adamreeve/npTDMS',
  packages = ['nptdms', 'nptdms.export', 'nptdms.test'],
  long_description=open('README.rst').read(),
  license = 'LGPL',
  classifiers = [
    'Development Status :: 4 - Beta',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Scientific/Engineering',
    'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
    'Intended Audience :: Science/Research',
    'Natural Language :: English',
  ],
  install_requires = ['numpy'],
  extras_require = {
      'test': [
          'pytest>=3.1.0',
          'hypothesis',
          'pytest-benchmark',
          'mock<4.0;python_version<"3.4"',
          'thermocouples_reference',
          'scipy',
      ],
      'pandas': ['pandas'],
      'hdf': ['h5py>=2.10.0'],
      'thermocouple_scaling': [],  # Kept for backwards compatibility
  },
  entry_points = """
  [console_scripts]
  tdmsinfo=nptdms.tdmsinfo:main
  """
)



"""Test TdmsSegment"""

from datetime import datetime
import pytest
try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = dict
import numpy as np

from nptdms.writer import ChannelObject, TdmsSegment, read_properties_dict
from nptdms.types import *


class StubObject(object):
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


def test_write_leadin_with_one_channel():
    data_type = StubObject(size=4)

    channel = StubObject(
        path="",
        has_data=True,
        data=[0] * 10,
        data_type=data_type)

    toc = ["kTocMetaData", "kTocRawData", "kTocNewObjList"]
    metadata_size = 12

    segment = TdmsSegment([channel])
    leadin = segment.leadin(toc, metadata_size)

    expected_values = [
        Bytes(b'TDSm'),
        Int32(14),  # TOC bitmask
        Int32(4712),  # TDMS version
        Uint64(52),  # Next segment offset
        Uint64(12),  # Raw data offset
        ]

    _assert_sequence_equal(leadin, expected_values)


def test_write_leadin_with_object_without_data():
    channel = StubObject(
        path="",
        has_data=False)

    toc = ["kTocMetaData", "kTocRawData", "kTocNewObjList"]
    metadata_size = 12

    segment = TdmsSegment([channel])
    leadin = segment.leadin(toc, metadata_size)

    expected_values = [
        Bytes(b'TDSm'),
        Int32(14),  # TOC bitmask
        Int32(4712),  # TDMS version
        Uint64(12),  # Next segment offset
        Uint64(12),  # Raw data offset
        ]

    _assert_sequence_equal(leadin, expected_values)


def test_write_metadata_with_properties():
    data_type = StubObject(enum_value=3)

    # Use an ordered dict for properties so that
    # the order of properties in metadata is guaranteed
    properties = OrderedDict()
    properties["prop1"] = String("foo")
    properties["prop2"] = Int32(42)

    channel = StubObject(
        path="channel_path",
        has_data=True,
        data=[1] * 10,
        data_type=data_type,
        properties=properties)

    segment = TdmsSegment([channel])
    metadata = segment.metadata()

    expected_values = [
        Uint32(1),  # Number of objects
        String("channel_path"),
        Uint32(20),  # Length of raw data index in bytes
        Int32(3),  # Data type
        Uint32(1),  # Array dimension
        Uint64(10),  # Number of values
        Uint32(2),  # Number of properties
        String("prop1"),  # Property name
        Int32(0x20),
        String("foo"),
        String("prop2"),
        Int32(3),
        Int32(42),
        ]

    _assert_sequence_equal(metadata, expected_values)


def test_write_metadata_with_no_data():
    obj = StubObject(
        path="object_path",
        has_data=False,
        properties={})

    segment = TdmsSegment([obj])
    metadata = segment.metadata()

    expected_values = [
        Uint32(1),  # Number of objects
        String("object_path"),
        Bytes(b'\xFF\xFF\xFF\xFF'),  # Raw data index
        Uint32(0),  # Number of properties
        ]

    _assert_sequence_equal(metadata, expected_values)


def test_properties_are_converted_to_tdms_types():
    test_time = datetime.utcnow()

    properties = {
        "prop1": Int32(1),
        "prop2": 2,
        "prop3": "foo",
        "prop4": True,
        "prop5": 3.142,
        "prop6": test_time,
    }

    tdms_properties = read_properties_dict(properties)

    assert len(tdms_properties) == len(properties)
    assert tdms_properties["prop1"] == Int32(1)
    assert tdms_properties["prop2"] == Int32(2)
    assert tdms_properties["prop3"] == String("foo")
    assert tdms_properties["prop4"] == Boolean(True)
    assert tdms_properties["prop5"] == DoubleFloat(3.142)
    assert tdms_properties["prop6"] == TimeStamp(test_time)


def test_datetime_converted_when_it_only_has_date_part():
    test_time = np.datetime64('2017-11-19')

    properties = {
        "time_prop": test_time,
    }

    tdms_properties = read_properties_dict(properties)

    assert tdms_properties["time_prop"] == TimeStamp(test_time)


def test_writing_long_integer_properties():
    properties = {
        "prop1": 2147483647,
        "prop2": 2147483648,
    }

    tdms_properties = read_properties_dict(properties)

    assert len(tdms_properties) == len(properties)
    assert tdms_properties["prop1"] == Int32(2147483647)
    assert tdms_properties["prop2"] == Int64(2147483648)


def test_writing_properties_with_numpy_typed_values():
    properties = {
        "int32prop": np.int32(32),
        "int64prop": np.int64(64),
        "float32prop": np.float32(32.0),
        "float64prop": np.float64(64.0),
        "boolprop": np.bool_(True),
    }

    tdms_properties = read_properties_dict(properties)

    assert len(tdms_properties) == len(properties)
    assert tdms_properties["int32prop"] == Int32(32)
    assert tdms_properties["int64prop"] == Int64(64)
    assert tdms_properties["float32prop"] == SingleFloat(32.0)
    assert tdms_properties["float64prop"] == DoubleFloat(64.0)
    assert tdms_properties["boolprop"] == Boolean(True)


def test_error_raised_when_cannot_convert_property_value():
    properties = {
        "prop1": None
    }

    with pytest.raises(TypeError):
        read_properties_dict(properties)


@pytest.mark.parametrize(
    "data",
    [
        np.array(0.0, dtype=np.dtype('float32')),
        np.zeros((2, 2), dtype=np.dtype('float32')),
        np.zeros((2, 2, 2), dtype=np.dtype('float32')),
    ]
)
def test_channel_with_invalid_data(data):
    with pytest.raises(ValueError) as exc_info:
        _ = ChannelObject("Group", "Channel", data, {})
    assert str(exc_info.value) == "Channel data must be a 1d array"


def _assert_sequence_equal(values, expected_values):
    position = 1
    expected_values = iter(expected_values)
    for val in values:
        try:
            expected = next(expected_values)
        except StopIteration:
            raise ValueError(
                "Expected end of sequence at position %d but found: %r" %
                (position, val))
        assert val == expected, "Expected %r to equal %r at position %d" % (val, expected, position)
        position += 1
    try:
        expected = next(expected_values)
        raise ValueError(
            "Expected %r at position %d but found end of sequence" %
            (expected, position))
    except StopIteration:
        pass



""" Python module for reading TDMS files produced by LabView

    This module contains the public facing API for reading TDMS files
"""

from collections import defaultdict
import numpy as np

from nptdms import scaling, types
from nptdms.utils import Timer, OrderedDict, cached_property
from nptdms.log import log_manager
from nptdms.common import ObjectPath
from nptdms.reader import TdmsReader
from nptdms.channel_data import get_data_receiver, slice_raw_data
from nptdms.export import hdf_export, pandas_export
from nptdms.base_segment import RawChannelDataChunk
from nptdms.timestamp import TdmsTimestamp, TimestampArray


log = log_manager.get_logger(__name__)


class TdmsFile(object):
    """ Reads and stores data from a TDMS file.

    There are two main ways to create a new TdmsFile object.
    TdmsFile.read will read all data into memory::

        tdms_file = TdmsFile.read(tdms_file_path)

    or you can use TdmsFile.open to read file metadata but not immediately read all data,
    for cases where a file is too large to easily fit in memory or you don't need to
    read data for all channels::

        with TdmsFile.open(tdms_file_path) as tdms_file:
            # Use tdms_file
            ...

    This class acts like a dictionary, where the keys are names of groups in the TDMS
    files and the values are TdmsGroup objects.
    A TdmsFile can be indexed by group name to access a group within the TDMS file, for example::

        tdms_file = TdmsFile.read(tdms_file_path)
        group = tdms_file[group_name]

    Iterating over a TdmsFile produces the names of groups in this file,
    or you can use the groups method to directly access all groups::

        for group in tdms_file.groups():
            # Use group
            ...
    """

    @staticmethod
    def read(file, raw_timestamps=False, memmap_dir=None):
        """ Creates a new TdmsFile object and reads all data in the file

        :param file: Either the path to the tdms file to read
            as a string or pathlib.Path, or an already opened file.
        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
            but this loses some precision.
            Setting this to true will read timestamps as a custom TdmsTimestamp type.
        :param memmap_dir: The directory to store memory mapped data files in,
            or None to read data into memory. The data files are created
            as temporary files and are deleted when the channel data is no
            longer used. tempfile.gettempdir() can be used to get the default
            temporary file directory.
        """
        return TdmsFile(file, raw_timestamps=raw_timestamps, memmap_dir=memmap_dir)

    @staticmethod
    def open(file, raw_timestamps=False, memmap_dir=None):
        """ Creates a new TdmsFile object and reads metadata, leaving the file open
            to allow reading channel data

        :param file: Either the path to the tdms file to read
            as a string or pathlib.Path, or an already opened file.
        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
            but this loses some precision.
            Setting this to true will read timestamps as a custom TdmsTimestamp type.
        :param memmap_dir: The directory to store memory mapped data files in,
            or None to read data into memory. The data files are created
            as temporary files and are deleted when the channel data is no
            longer used. tempfile.gettempdir() can be used to get the default
            temporary file directory.
        """
        return TdmsFile(
            file, raw_timestamps=raw_timestamps, memmap_dir=memmap_dir, read_metadata_only=True, keep_open=True)

    @staticmethod
    def read_metadata(file, raw_timestamps=False):
        """ Creates a new TdmsFile object and only reads the metadata

        :param file: Either the path to the tdms file to read
            as a string or pathlib.Path, or an already opened file.
        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
            but this loses some precision.
            Setting this to true will read timestamps as a custom TdmsTimestamp type.
        """
        return TdmsFile(file, raw_timestamps=raw_timestamps, read_metadata_only=True)

    def __init__(self, file, raw_timestamps=False, memmap_dir=None, read_metadata_only=False, keep_open=False):
        """Initialise a new TdmsFile object

        :param file: Either the path to the tdms file to read
            as a string or pathlib.Path, or an already opened file.
        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
            but this loses some precision.
            Setting this to true will read timestamps as a custom TdmsTimestamp type.
        :param memmap_dir: The directory to store memory mapped data files in,
            or None to read data into memory. The data files are created
            as temporary files and are deleted when the channel data is no
            longer used. tempfile.gettempdir() can be used to get the default
            temporary file directory.
        :param read_metadata_only: If this parameter is enabled then only the
            metadata of the TDMS file will read.
        :param keep_open: Keeps the file open so data can be read if only metadata
            is read initially.
        """

        self._memmap_dir = memmap_dir
        self._raw_timestamps = raw_timestamps
        self._groups = OrderedDict()
        self._properties = OrderedDict()
        self._channel_data = {}
        self.data_read = False

        self._reader = TdmsReader(file)
        try:
            self._read_file(self._reader, read_metadata_only, keep_open)
        finally:
            if not keep_open:
                self._reader.close()

    def groups(self):
        """Returns a list of the groups in this file

        :rtype: List of TdmsGroup.
        """

        return list(self._groups.values())

    @property
    def properties(self):
        """ Return the properties of this file as a dictionary

        These are the properties associated with the root TDMS object.
        """

        return self._properties

    def as_dataframe(self, time_index=False, absolute_time=False, scaled_data=True):
        """
        Converts the TDMS file to a DataFrame. DataFrame columns are named using the TDMS object paths.

        :param time_index: Whether to include a time index for the dataframe.
        :param absolute_time: If time_index is true, whether the time index
            values are absolute times or relative to the start time.
        :param scaled_data: By default the scaled data will be used.
            Set to False to use raw unscaled data.
            For DAQmx data, there will be one column per DAQmx raw scaler and column names will include the scale id.
        :return: The full TDMS file data.
        :rtype: pandas.DataFrame
        """

        return pandas_export.from_tdms_file(self, time_index, absolute_time, scaled_data)

    def as_hdf(self, filepath, mode='w', group='/'):
        """
        Converts the TDMS file into an HDF5 file

        :param filepath: The path of the HDF5 file you want to write to.
        :param mode: The write mode of the HDF5 file. This can be 'w' or 'a'
        :param group: A group in the HDF5 file that will contain the TDMS data.
        """
        return hdf_export.from_tdms_file(self, filepath, mode, group)

    def data_chunks(self):
        """ A generator that streams chunks of data from disk.
        This method may only be used when the TDMS file was opened without reading all data immediately.

        :rtype: Generator that yields :class:`DataChunk` objects
        """
        channel_offsets = defaultdict(int)
        for chunk in self._reader.read_raw_data():
            _convert_data_chunk(chunk, self._raw_timestamps)
            yield DataChunk(self, chunk, channel_offsets)
            for path, data in chunk.channel_data.items():
                channel_offsets[path] += len(data)

    def close(self):
        """ Close the underlying file if it was opened by this TdmsFile

            If this TdmsFile was initialised with an already open file
            then the reference to it is released but the file is not closed.
        """
        if self._reader is not None:
            self._reader.close()
            self._reader = None

    def __len__(self):
        """ Returns the number of groups in this file
        """
        return len(self._groups)

    def __iter__(self):
        """ Returns an iterator over the names of groups in this file
        """
        return iter(self._groups)

    def __getitem__(self, group_name):
        """ Retrieve a TDMS group from the file by name
        """
        try:
            return self._groups[group_name]
        except KeyError:
            raise KeyError("There is no group named '%s' in the TDMS file" % group_name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _read_file(self, tdms_reader, read_metadata_only, keep_open):
        tdms_reader.read_metadata(require_segment_indexes=keep_open)

        # Use object metadata to build group and channel objects
        group_properties = OrderedDict()
        group_channels = OrderedDict()
        object_properties = {
            path_string: self._convert_properties(obj.properties)
            for path_string, obj in tdms_reader.object_metadata.items()}
        try:
            self._properties = object_properties['/']
        except KeyError:
            pass

        for (path_string, obj) in tdms_reader.object_metadata.items():
            properties = object_properties[path_string]
            path = ObjectPath.from_string(path_string)
            if path.is_root:
                pass
            elif path.is_group:
                group_properties[path.group] = properties
            else:
                # Object is a channel
                try:
                    channel_group_properties = object_properties[path.group_path()]
                except KeyError:
                    channel_group_properties = OrderedDict()
                channel = TdmsChannel(
                    path, obj.data_type, obj.scaler_data_types, obj.num_values,
                    properties, channel_group_properties, self._properties,
                    tdms_reader, self._raw_timestamps, self._memmap_dir)
                if path.group in group_channels:
                    group_channels[path.group].append(channel)
                else:
                    group_channels[path.group] = [channel]

        # Create group objects containing channels and properties
        for group_name, properties in group_properties.items():
            try:
                channels = group_channels[group_name]
            except KeyError:
                channels = []
            group_path = ObjectPath(group_name)
            self._groups[group_name] = TdmsGroup(group_path, properties, channels)
        for group_name, channels in group_channels.items():
            if group_name not in self._groups:
                # Group with channels but without any corresponding object metadata in the file:
                group_path = ObjectPath(group_name)
                self._groups[group_name] = TdmsGroup(group_path, {}, channels)

        if not read_metadata_only:
            self._read_data(tdms_reader)

    def _read_data(self, tdms_reader):
        with Timer(log, "Allocate space"):
            # Allocate space for data
            for group in self.groups():
                for channel in group.channels():
                    self._channel_data[channel.path] = get_data_receiver(
                        channel, len(channel), self._raw_timestamps, self._memmap_dir)

        with Timer(log, "Read data"):
            # Now actually read all the data
            for chunk in tdms_reader.read_raw_data():
                for (path, data) in chunk.channel_data.items():
                    channel_data = self._channel_data[path]
                    if data.data is not None:
                        channel_data.append_data(data.data)
                    elif data.scaler_data is not None:
                        for scaler_id, scaler_data in data.scaler_data.items():
                            channel_data.append_scaler_data(scaler_id, scaler_data)

            for group in self.groups():
                for channel in group.channels():
                    channel_data = self._channel_data[channel.path]
                    if channel_data is not None:
                        channel._set_raw_data(channel_data)

        self.data_read = True

    def _convert_properties(self, properties):
        def convert_prop(val):
            if isinstance(val, TdmsTimestamp) and not self._raw_timestamps:
                # Convert timestamps to numpy datetime64 if raw timestamps are not requested
                return val.as_datetime64()
            return val
        return OrderedDict((k, convert_prop(v)) for (k, v) in properties.items())


class TdmsGroup(object):
    """ Represents a group of channels in a TDMS file.

    This class acts like a dictionary, where the keys are names of channels in the group
    and the values are TdmsChannel objects.
    A TdmsGroup can be indexed by channel name to access a channel in this group, for example::

        channel = group[channel_name]

    Iterating over a TdmsGroup produces the names of channels in this group,
    or you can use the channels method to directly access all channels::

        for channel in group.channels():
            # Use channel
            ...

    :ivar ~.properties: Dictionary of TDMS properties defined for this group.
    """

    def __init__(self, path, properties, channels):
        self._path = path
        self.properties = properties
        self._channels = {c.name: c for c in channels}

    def __repr__(self):
        return "<TdmsGroup with path %s>" % self.path

    @property
    def path(self):
        """ Path to the TDMS object for this group
        """
        return str(self._path)

    @property
    def name(self):
        """ The name of this group
        """
        return self._path.group

    def channels(self):
        """ The list of channels in this group

        :rtype: A list of TdmsChannel
        """
        return list(self._channels.values())

    def as_dataframe(self, time_index=False, absolute_time=False, scaled_data=True):
        """
        Converts the TDMS group to a DataFrame. DataFrame columns are named using the channel names.

        :param time_index: Whether to include a time index for the dataframe.
        :param absolute_time: If time_index is true, whether the time index
            values are absolute times or relative to the start time.
        :param scaled_data: By default the scaled data will be used.
            Set to False to use raw unscaled data.
            For DAQmx data, there will be one column per DAQmx raw scaler and column names will include the scale id.
        :return: The TDMS object data.
        :rtype: pandas.DataFrame
        """

        return pandas_export.from_group(self, time_index, absolute_time, scaled_data)

    def __len__(self):
        """ Returns the number of channels in this group
        """
        return len(self._channels)

    def __iter__(self):
        """ Returns an iterator over the names of channels in this group
        """
        return iter(self._channels)

    def __getitem__(self, channel_name):
        """ Retrieve a TDMS channel from this group by name
        """
        try:
            return self._channels[channel_name]
        except KeyError:
            raise KeyError(
                "There is no channel named '%s' in group '%s' of the TDMS file" %
                (channel_name, self.name))


class TdmsChannel(object):
    """ Represents a data channel in a TDMS file.

    This class acts like an array, you can get the length of a channel using :code:`len(channel)`,
    and can iterate over values in the channel using a for loop,
    or index into a channel using an integer index to get a single value::

        for value in channel:
            # Use value
            ...
        first_value = channel[0]

    Or you can index using a slice to retrieve a range of data as a numpy array.
    To get all data in this channel as a numpy array::

        all_data = channel[:]

    Or to retrieve a subset of data::

        data_subset = channel[start:stop]

    :ivar ~.properties: Dictionary of TDMS properties defined for this channel,
                      for example the start time and time increment for waveforms.
    """

    def __init__(
            self, path, data_type, scaler_data_types, number_values,
            properties, group_properties, file_properties,
            tdms_reader, raw_timestamps, memmap_dir):
        self._path = path
        self.properties = properties
        self._length = number_values
        self.data_type = data_type
        self.scaler_data_types = scaler_data_types
        self._group_properties = group_properties
        self._file_properties = file_properties
        self._reader = tdms_reader
        self._raw_timestamps = raw_timestamps
        self._memmap_dir = memmap_dir

        self._raw_data = None
        self._cached_chunk = None
        self._cached_chunk_bounds = None

    def __repr__(self):
        return "<TdmsChannel with path %s>" % self.path

    def __len__(self):
        """ Returns the number of values in this channel
        """
        return self._length

    def __iter__(self):
        """ Returns an iterator over the values in this channel
        """
        if self._raw_data is not None:
            return iter(self.data)
        else:
            return self._read_data_values()

    def __getitem__(self, index):
        if self._raw_data is not None:
            return self.data[index]
        elif index is Ellipsis:
            return self.read_data()
        elif isinstance(index, slice):
            return self._read_slice(index.start, index.stop, index.step)
        elif isinstance(index, int):
            return self._read_at_index(index)
        else:
            raise TypeError("Invalid index type '%s', expected int, slice or Ellipsis" % type(index).__name__)

    @property
    def path(self):
        """ Path to the TDMS object for this channel
        """
        return str(self._path)

    @property
    def name(self):
        """ The name of this channel
        """
        return self._path.channel

    @property
    def group_name(self):
        """ The name of the group that contains this channel
        """
        return self._path.group

    @cached_property
    def dtype(self):
        """ NumPy data type of the channel data

        For data with a scaling this is the data type of the scaled data

        :rtype: numpy.dtype
        """
        channel_scaling = self._scaling
        if channel_scaling is not None:
            return channel_scaling.get_dtype(self.data_type, self.scaler_data_types)
        return self._raw_data_dtype()

    def _raw_data_dtype(self):
        if self.data_type is types.String:
            return np.dtype('O')
        elif self.data_type is types.TimeStamp:
            return np.dtype('<M8[us]')
        if self.data_type is not None and self.data_type.nptype is not None:
            return self.data_type.nptype
        return np.dtype('V8')

    @cached_property
    def data(self):
        """ If the TdmsFile was created by reading all data, this property
        provides direct access to the numpy array containing the data for this channel.

        Indexing into the channel with a slice should be preferred to using this property, for example::

            channel_data = channel[:]
        """
        if len(self) > 0 and self._raw_data is None:
            raise RuntimeError("Channel data has not been read")

        if self._raw_data is None:
            return np.empty((0, ), dtype=self.dtype)
        return self._scale_data(self._raw_data)

    @property
    def raw_data(self):
        """ If the TdmsFile was created by reading all data, this property
        provides direct access to the numpy array of raw, unscaled data.
        For unscaled objects this is the same as the data property.
        """
        if len(self) > 0 and self._raw_data is None:
            raise RuntimeError("Channel data has not been read")

        if self._raw_data is None:
            return np.empty((0, ), dtype=self._raw_data_dtype())
        if self._raw_data.scaler_data:
            if len(self._raw_data.scaler_data) == 1:
                return next(v for v in self._raw_data.scaler_data.values())
            else:
                raise Exception(
                    "This object has data for multiple DAQmx scalers, "
                    "use the raw_scaler_data property to get raw data "
                    "for a scale_id")
        return self._raw_data.data

    @property
    def raw_scaler_data(self):
        """ If the TdmsFile was created by reading all data, this property
        provides direct access to the numpy array of raw DAQmx scaler data
        as a dictionary mapping from scale id to raw data arrays.
        """
        if len(self) > 0 and self._raw_data is None:
            raise RuntimeError("Channel data has not been read")

        return self._raw_data.scaler_data

    def data_chunks(self):
        """ A generator that streams chunks data for this channel from disk.
        This method may only be used when the TDMS file was opened without reading all data immediately.

        :rtype: Generator that yields :class:`ChannelDataChunk` objects
        """
        channel_offset = 0
        for raw_data_chunk in self._read_channel_data_chunks():
            yield ChannelDataChunk(self, raw_data_chunk, channel_offset)
            channel_offset += len(raw_data_chunk)

    def read_data(self, offset=0, length=None, scaled=True):
        """ Reads data for this channel from the TDMS file and returns it as a numpy array

        Indexing into the channel with a slice should be preferred over using
        this method, but this method is needed if you want to read raw, unscaled data.

        :param offset: Initial position to read data from.
        :param length: Number of values to attempt to read.
            Fewer values will be returned if attempting to read beyond the end of the available data.
        :param scaled: By default scaling will be applied to the returned data.
            Set this parameter to False to return raw unscaled data.
            For DAQmx data a dictionary of scaler id to raw scaler data will be returned.
        """
        if self._raw_data is None:
            raw_data = self._read_channel_data(offset, length)
        else:
            raw_data = slice_raw_data(self._raw_data, offset, length)

        if raw_data is None:
            dtype = self.dtype if scaled else self._raw_data_dtype()
            return np.empty((0,), dtype=dtype)
        if scaled:
            return self._scale_data(raw_data)
        else:
            if raw_data.scaler_data:
                return raw_data.scaler_data
            return raw_data.data

    def time_track(self, absolute_time=False, accuracy='ns'):
        """Return an array of time or the independent variable for this channel

        This depends on the object having the wf_increment
        and wf_start_offset properties defined.
        Note that wf_start_offset is usually zero for time-series data.
        If you have time-series data channels with different start times,
        you should use the absolute time or calculate the time offsets using
        the wf_start_time property.

        For larger timespans, the accuracy setting should be set lower.
        The default setting is 'ns', which has a timespan of
        [1678 AD, 2262 AD]. For the exact ranges, refer to
        http://docs.scipy.org/doc/numpy/reference/arrays.datetime.html
        section "Datetime Units".

        :param absolute_time: Whether the returned time values are absolute
            times rather than relative to the start time. If true, the
            wf_start_time property must be set.
        :param accuracy: The accuracy of the returned datetime64 array.
        :rtype: NumPy array.
        :raises: KeyError if required properties aren't found

        """

        try:
            increment = self.properties['wf_increment']
            offset = self.properties['wf_start_offset']
        except KeyError:
            raise KeyError("Object does not have time properties available.")

        relative_time = np.linspace(
            offset,
            offset + (len(self) - 1) * increment,
            len(self))

        if not absolute_time:
            return relative_time

        try:
            start_time = self.properties['wf_start_time']
        except KeyError:
            raise KeyError(
                "Object does not have start time property available.")

        try:
            unit_correction = {
                's': 1e0,
                'ms': 1e3,
                'us': 1e6,
                'ns': 1e9,
            }[accuracy]
        except KeyError:
            raise KeyError("Invalid accuracy: {0}".format(accuracy))

        # Because numpy only knows ints as its date datatype,
        # convert to accuracy.
        time_type = "timedelta64[{0}]".format(accuracy)
        return (np.datetime64(start_time) +
                (relative_time * unit_correction).astype(time_type))

    def as_dataframe(self, time_index=False, absolute_time=False, scaled_data=True):
        """
        Converts the TDMS channel to a DataFrame. The DataFrame column is named using the channel path.

        :param time_index: Whether to include a time index for the dataframe.
        :param absolute_time: If time_index is true, whether the time index
            values are absolute times or relative to the start time.
        :param scaled_data: By default the scaled data will be used.
            Set to False to use raw unscaled data.
            For DAQmx data, there will be one column per DAQmx raw scaler and column names will include the scale id.
        :return: The TDMS object data.
        :rtype: pandas.DataFrame
        """

        return pandas_export.from_channel(self, time_index, absolute_time, scaled_data)

    def _read_data_values(self):
        for chunk in self.data_chunks():
            for value in chunk:
                yield value

    def _read_slice(self, start, stop, step):
        if step == 0:
            raise ValueError("Step size cannot be zero")

        # Replace None values with defaults
        step = 1 if step is None else step
        if start is None:
            start = 0 if step > 0 else -1
        if stop is None:
            stop = self._length if step > 0 else -1 - self._length

        # Handle negative indices
        if start < 0:
            start = self._length + start
        if stop < 0:
            stop = self._length + stop

        # Check for empty ranges
        if stop == start:
            return np.empty((0, ), dtype=self.dtype)
        if step > 0 and (stop < start or start >= self._length or stop < 0):
            return np.empty((0,), dtype=self.dtype)
        if step < 0 and (stop > start or stop >= self._length or start < 0):
            return np.empty((0,), dtype=self.dtype)

        # Trim values outside bounds
        if start < 0:
            start = 0
        if start >= self._length:
            start = self._length - 1
        if stop > self._length:
            stop = self._length
        if stop < -1:
            stop = -1

        # Read data and handle step size
        if step > 0:
            read_data = self.read_data(start, stop - start)
            return read_data[::step] if step > 1 else read_data
        else:
            read_data = self.read_data(stop + 1, start - stop)
            return read_data[::step]

    def _read_at_index(self, index):
        orig_index = index
        if index < 0:
            index = self._length + index
        if index < 0 or index >= self._length:
            raise IndexError("Index {0} is outside of the channel bounds [0, {1}]".format(orig_index, self._length - 1))

        if self._cached_chunk is not None:
            # Check if we've already read and cached the chunk containing this index
            bounds = self._cached_chunk_bounds
            if bounds[0] <= index < bounds[1]:
                return self._cached_chunk[index - bounds[0]]

        chunk, chunk_offset = self._read_channel_data_chunk_for_index(index)
        scaled_chunk = self._scale_data(chunk)
        self._cached_chunk = scaled_chunk
        self._cached_chunk_bounds = (chunk_offset, chunk_offset + len(scaled_chunk))

        return scaled_chunk[index - chunk_offset]

    def _scale_data(self, raw_data):
        scale = self._scaling
        if scale is not None:
            return scale.scale(raw_data)
        elif raw_data.scaler_data:
            raise ValueError("Missing scaling information for DAQmx data")
        else:
            return raw_data.data

    @cached_property
    def _scaling(self):
        return scaling.get_scaling(
            self.properties, self._group_properties, self._file_properties)

    def _read_channel_data_chunks(self):
        for chunk in self._reader.read_raw_data_for_channel(self.path):
            _convert_channel_data_chunk(chunk, self._raw_timestamps)
            yield chunk

    def _read_channel_data_chunk_for_index(self, index):
        (chunk, offset) = self._reader.read_channel_chunk_for_index(self.path, index)
        _convert_channel_data_chunk(chunk, self._raw_timestamps)
        return chunk, offset

    def _read_channel_data(self, offset=0, length=None):
        if offset < 0:
            raise ValueError("offset must be non-negative")
        if length is not None and length < 0:
            raise ValueError("length must be non-negative")

        with Timer(log, "Allocate space for channel"):
            # Allocate space for data
            if length is None:
                num_values = len(self) - offset
            else:
                num_values = min(length, len(self) - offset)
            num_values = max(0, num_values)
            channel_data = get_data_receiver(self, num_values, self._raw_timestamps, self._memmap_dir)

        with Timer(log, "Read data for channel"):
            # Now actually read all the data
            for chunk in self._reader.read_raw_data_for_channel(self.path, offset, length):
                if chunk.data is not None:
                    channel_data.append_data(chunk.data)
                if chunk.scaler_data is not None:
                    for scaler_id, scaler_data in chunk.scaler_data.items():
                        channel_data.append_scaler_data(scaler_id, scaler_data)

        return channel_data

    def _set_raw_data(self, data):
        self._raw_data = data


class DataChunk(object):
    """ A chunk of data in a TDMS file

    Can be indexed by group name to get the data for a group in this channel,
    which can then be indexed by channel name to get the data for a channel in this chunk.
    For example::

        group_chunk = data_chunk[group_name]
        channel_chunk = group_chunk[channel_name]
    """
    def __init__(self, tdms_file, raw_data_chunk, channel_offsets):
        self._groups = OrderedDict(
            (group.name, GroupDataChunk(tdms_file, group, raw_data_chunk, channel_offsets))
            for group in tdms_file.groups())

    def __getitem__(self, group_name):
        """ Get a chunk of data for a group
        """
        return self._groups[group_name]

    def groups(self):
        """ Returns chunks of data for all groups

        :rtype: List of :class:`GroupDataChunk`
        """
        return list(self._groups.values())


class GroupDataChunk(object):
    """ A chunk of data for a group in a TDMS file

    Can be indexed by channel name to get the data for a channel in this chunk.
    For example::

        channel_chunk = group_chunk[channel_name]

    :ivar ~.name: Name of the group
    """
    def __init__(self, tdms_file, group, raw_data_chunk, channel_offsets):
        self.name = group.name
        self._channels = OrderedDict(
            (channel.name, ChannelDataChunk(
                channel,
                raw_data_chunk.channel_data.get(channel.path, RawChannelDataChunk.empty()),
                channel_offsets[channel.path]))
            for channel in group.channels())

    def __getitem__(self, channel_name):
        """ Get a chunk of data for a channel in this group
        """
        return self._channels[channel_name]

    def channels(self):
        """ Returns chunks of channel data for all channels in this group

        :rtype: List of :class:`ChannelDataChunk`
        """
        return list(self._channels.values())


class ChannelDataChunk(object):
    """ A chunk of data for a channel in a TDMS file

    Is an array-like object that supports indexing to access data, for example::

        chunk_length = len(channel_data_chunk)
        chunk_data = channel_data_chunk[:]

    :ivar ~.name: Name of the channel
    :ivar ~.offset: Starting index of this chunk of data in the entire channel
    """
    def __init__(self, channel, raw_data_chunk, offset):
        self._path = channel._path
        self._channel = channel
        self.name = channel.name
        self.offset = offset
        self._raw_data = raw_data_chunk

    def __len__(self):
        """ Returns the number of values in this chunk
        """
        return len(self._raw_data)

    def __getitem__(self, index):
        """ Get a value or slice of values from this chunk
        """
        return self._data()[index]

    def __iter__(self):
        """ Iterate over values in this chunk
        """
        return iter(self._data())

    def _data(self):
        if self._raw_data.data is None and self._raw_data.scaler_data is None:
            return np.empty((0, ), dtype=self._channel.dtype)

        scale = self._channel._scaling
        if scale is not None:
            return scale.scale(self._raw_data)
        elif self._raw_data.scaler_data:
            raise ValueError("Missing scaling information for DAQmx data")
        else:
            return self._raw_data.data


def _convert_data_chunk(chunk, raw_timestamps):
    for channel_chunk in chunk.channel_data.values():
        _convert_channel_data_chunk(channel_chunk, raw_timestamps)


def _convert_channel_data_chunk(channel_chunk, raw_timestamps):
    if not raw_timestamps and isinstance(channel_chunk.data, TimestampArray):
        channel_chunk.data = channel_chunk.data.as_datetime64()



""" Test reading timestamp properties and data
"""

from datetime import datetime
import numpy as np
import struct
import pytest
from nptdms import TdmsFile
from nptdms.timestamp import TdmsTimestamp, TimestampArray
from nptdms.types import TimeStamp
from nptdms.test.util import (
    GeneratedFile,
    channel_metadata,
    hexlify_value,
    segment_objects_metadata,
)


def test_read_raw_timestamp_properties():
    """ Test reading timestamp properties as a raw TDMS timestamp
    """
    test_file = GeneratedFile()
    second_fractions = 1234567890 * 10 ** 10
    properties = {
        "wf_start_time": (0x44, hexlify_value("<Q", second_fractions) + hexlify_value("<q", 3524551547))
    }
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 3, 2, properties),
        ),
        "01 00 00 00" "02 00 00 00"
    )

    with test_file.get_tempfile() as temp_file:
        tdms_data = TdmsFile.read(temp_file.file, raw_timestamps=True)
        start_time = tdms_data['group']['channel1'].properties['wf_start_time']
        assert start_time.seconds == 3524551547
        assert start_time.second_fractions == second_fractions


def test_timestamp_as_datetime64():
    """ Test converting a timestamp to a numpy datetime64
    """
    second_fractions = 1234567890 * 10 ** 10
    seconds = 3524551547
    timestamp = TdmsTimestamp(seconds, second_fractions)

    assert timestamp.as_datetime64() == np.datetime64('2015-09-08T10:05:47.669260', 'us')
    assert timestamp.as_datetime64().dtype == np.dtype('datetime64[us]')
    assert timestamp.as_datetime64('ns') == np.datetime64('2015-09-08T10:05:47.669260594', 'ns')
    assert timestamp.as_datetime64('ns').dtype == np.dtype('datetime64[ns]')


def test_timestamp_as_datetime():
    """ Test converting a timestamp to a datetime.datetime
    """
    second_fractions = 1234567890 * 10 ** 10
    seconds = 3524551547
    timestamp = TdmsTimestamp(seconds, second_fractions)

    assert timestamp.as_datetime() == datetime(2015, 9, 8, 10, 5, 47, 669261)


def test_read_raw_timestamp_data():
    """ Test reading timestamp data as a raw TDMS timestamps
    """
    test_file = GeneratedFile()
    seconds = 3672033330
    second_fractions = 1234567890 * 10 ** 10
    test_file.add_segment(
        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
        segment_objects_metadata(
            channel_metadata("/'group'/'channel1'", 0x44, 4),
        ),
        hexlify_value("<Q", 0) + hexlify_value("<q", seconds) +
        hexlify_value("<Q", second_fractions) + hexlify_value("<q", seconds) +
        hexlify_value("<Q", 0) + hexlify_value("<q", seconds + 1) +
        hexlify_value("<Q", second_fractions) + hexlify_value("<q", seconds + 1)
    )

    expected_seconds = np.array([seconds, seconds, seconds + 1, seconds + 1], np.dtype('int64'))
    expected_second_fractions = np.array([0, second_fractions, 0, second_fractions], np.dtype('uint64'))

    with test_file.get_tempfile() as temp_file:
        tdms_data = TdmsFile.read(temp_file.file, raw_timestamps=True)
        data = tdms_data['group']['channel1'][:]
        assert isinstance(data, TimestampArray)
        np.testing.assert_equal(data.seconds, expected_seconds)
        np.testing.assert_equal(data.second_fractions, expected_second_fractions)


def test_read_big_endian_timestamp_data():
    seconds = 3672033330
    second_fractions = 1234567890 * 10 ** 10
    data = (
        struct.pack(">q", seconds) + struct.pack(">Q", 0) +
        struct.pack(">q", seconds) + struct.pack(">Q", second_fractions) +
        struct.pack(">q", seconds + 1) + struct.pack(">Q", 0) +
        struct.pack(">q", seconds + 1) + struct.pack(">Q", second_fractions))
    expected_seconds = np.array([seconds, seconds, seconds + 1, seconds + 1], np.dtype('int64'))
    expected_second_fractions = np.array([0, second_fractions, 0, second_fractions], np.dtype('uint64'))

    timestamp_array = TimeStamp.from_bytes(np.frombuffer(data, dtype=np.dtype('uint8')), '>')

    np.testing.assert_equal(timestamp_array.seconds, expected_seconds)
    np.testing.assert_equal(timestamp_array.second_fractions, expected_second_fractions)


def test_timestamp_repr():
    timestamp = TdmsTimestamp(3672033330, 12345678900000000000)
    assert repr(timestamp) == 'TdmsTimestamp(3672033330, 12345678900000000000)'


def test_timestamp_str():
    timestamp = TdmsTimestamp(3672033330, 12345678900000000000)
    assert str(timestamp) == '2020-05-11T09:15:30.669261'


def test_timestamp_array_slicing():
    timestamp_array = _get_test_timestamp_array()
    array_slice = timestamp_array[0:2]
    assert isinstance(array_slice, TimestampArray)


def test_timestamp_array_get_single_item():
    timestamp_array = _get_test_timestamp_array()
    array_item = timestamp_array[3]
    assert isinstance(array_item, TdmsTimestamp)
    assert array_item.seconds == 3672033331
    assert array_item.second_fractions == 1234567890 * 10 ** 10


def test_timestamp_array_field_access():
    timestamp_array = _get_test_timestamp_array()
    seconds = timestamp_array.seconds
    second_fractions = timestamp_array.second_fractions
    assert isinstance(seconds, np.ndarray)
    assert not isinstance(seconds, TimestampArray)
    assert seconds.dtype == np.dtype('<i8')
    assert isinstance(second_fractions, np.ndarray)
    assert not isinstance(second_fractions, TimestampArray)
    assert second_fractions.dtype == np.dtype('<u8')


def test_timestamp_array_to_datetime64():
    timestamp_array = _get_test_timestamp_array()
    expected_timestamps = np.array([
        np.datetime64('2020-05-11 09:15:30'),
        np.datetime64('2020-05-11 09:15:30.669260'),
        np.datetime64('2020-05-11 09:15:31'),
        np.datetime64('2020-05-11 09:15:31.669260'),
    ])

    us_array = timestamp_array.as_datetime64()

    np.testing.assert_equal(us_array, expected_timestamps)


def test_timestamp_array_to_datetime64_with_ns_precision():
    timestamp_array = _get_test_timestamp_array()
    expected_timestamps = np.array([
        np.datetime64('2020-05-11 09:15:30'),
        np.datetime64('2020-05-11 09:15:30.669260594'),
        np.datetime64('2020-05-11 09:15:31'),
        np.datetime64('2020-05-11 09:15:31.669260594'),
    ])

    ns_array = timestamp_array.as_datetime64('ns')

    np.testing.assert_equal(ns_array, expected_timestamps)


@pytest.mark.parametrize(
    "input_array",
    [
        np.array([1, 2, 3, 4]),
        np.array([(1, 2), (3, 4)], dtype=[('a', '<i8'), ('b', '<u8')]),
        np.array([(1, 2), (3, 4)], dtype=[('seconds', '<i8'), ('b', '<u8')]),
        np.array([(1, 2), (3, 4)], dtype=[('a', '<i8'), ('second_fractions', '<u8')]),
    ]
)
def test_error_raised_with_creating_timestamp_array_with_invalid_input_type(input_array):
    with pytest.raises(ValueError) as exc_info:
        _ = TimestampArray(input_array)
    assert str(exc_info.value) == "Input array must have a dtype with 'seconds' and 'second_fractions' fields"


def test_error_raised_converting_timestamp_with_invalid_resolution():
    timestamp_array = _get_test_timestamp_array()
    timestamp = timestamp_array[0]
    with pytest.raises(ValueError) as exc_info:
        _ = timestamp.as_datetime64('invalid_res')
    assert str(exc_info.value) == "Unsupported resolution for converting to numpy datetime64: 'invalid_res'"


def test_error_raised_converting_timestamp_array_with_invalid_resolution():
    timestamp_array = _get_test_timestamp_array()
    with pytest.raises(ValueError) as exc_info:
        _ = timestamp_array.as_datetime64('invalid_res')
    assert str(exc_info.value) == "Unsupported resolution for converting to numpy datetime64: 'invalid_res'"


def _get_test_timestamp_array():
    dtype = np.dtype([('seconds', '<i8'), ('second_fractions', '<u8')])
    seconds = 3672033330
    second_fractions = 1234567890 * 10 ** 10
    array = np.array([
        (seconds, 0),
        (seconds, second_fractions),
        (seconds + 1, 0),
        (seconds + 1, second_fractions),
    ], dtype=dtype)
    return TimestampArray(array)
